@@ -1,30 +1,28 @@
 """
 NBX-Build Instances are APIs to your machines. These APIs can be used to change state of the machine (start, stop,
-etc.), can be used to transfer files to and from the machine (using ``Instance.mv`` commands), calling any shell
-command using ``Instance.remote``.
+etc.), can be used to transfer files to and from the machine (using `Instance.mv` commands), calling any shell
+command using `Instance.remote`.
 
-CLI Commands
-------------
+### CLI Commands
 
 Build comes built in with several functions to move and list files from the instance. All you need to know is that
-cloud files have prefix ``nbx://`` which is where your projects are. Here's a quick list:
+cloud files have prefix `nbx://` which is where your projects are. Here's a quick list:
 
-.. code-block:: bash
+```bash
 
-  # move files
-  nbx build -i 'instance_name' --workspace_id 'workspace_id' \
-    mv ./local_file nbx://cloud_file
+# move files
+nbx build -i 'instance_name' --workspace_id 'workspace_id' \
+  mv ./local_file nbx://cloud_file
 
-  # or move folders
-  nbx build -i 'instance_name' --workspace_id 'workspace_id' \
-    mv ./local_folder nbx://in_this/folder/
+# or move folders
+nbx build -i 'instance_name' --workspace_id 'workspace_id' \
+  mv ./local_folder nbx://in_this/folder/
 
+```
 
-You might be required to agree to the SSH connection being setup. If you want to avoid that set ``NBOX_SSH_NO_HOST_CHECKING=1``.
+You might be required to agree to the SSH connection being setup. If you want to avoid that set `NBOX_SSH_NO_HOST_CHECKING=1`.
 All these APIs are also available in python.
 
-Documentation
--------------
 """
 
 import io
@@ -38,9 +36,9 @@ from typing import Dict, List
 from requests.sessions import Session
 
 import nbox.utils as U
-from nbox.auth import secret
+from nbox.auth import secret, AuthConfig
 from nbox.utils import logger
-from nbox.subway import TIMEOUT_CALLS
+from nbox.subway import TIMEOUT_CALLS, Sub30
 from nbox.init import nbox_ws_v1, create_webserver_subway
 
 
@@ -53,16 +51,18 @@ make the entire process functional.
 '''
 ################################################################################
 
-def print_status(workspace_id: str = None, fields: List[str] = None):
-  """Print complete status of NBX-Build instances. If ``workspace_id`` is not provided
-  personal workspace will be used. Used in CLI"""
+def print_status(fields: List[str] = [], *, workspace_id: str = ""):
+  """Print complete status of NBX-Build instances. If `workspace_id` is not provided
+  personal workspace will be used. Used in CLI
+  
+  Args:
+    fields (List[str], optional): fields to print. Defaults to []. If not provided all fields will be printed.
+  """
   logger.info("Getting NBX-Build details")
-  if workspace_id == None:
-    stub_projects = nbox_ws_v1.user.projects
-  else:
-    stub_projects = nbox_ws_v1.workspace.u(workspace_id).projects
+  workspace_id = workspace_id or secret(AuthConfig.workspace_id)
+  stub_projects = nbox_ws_v1.projects
 
-  fields = fields if fields != None else Instance.useful_keys
+  fields = fields or Instance.useful_keys
 
   data = stub_projects()
   projects = data["project_details"]
@@ -86,26 +86,21 @@ class Instance():
   # each instance has a lot of data against it, we need to store only a few as attributes
   useful_keys = ["project_id", "project_name", "size_used", "size", "state"]
 
-  def __init__(self, i: str, workspace_id: int = None):
-    """NBX-Build Instance class manages the both individual instance, but provides
-    webserver functionality using ``nbox_ws_v1``, such as starting and stopping,
-    deletion and more.
+  def __init__(self, i: str, *, workspace_id: str = ""):
+    """NBX-Build Instance class manages the both individual instance, but provides webserver functionality using
+    `nbox_ws_v1`, such as starting and stopping, deletion and more.
 
     Args:
-        i (str): name or ``project_id`` of the instance
-        workspace_id (int, optional): id of the workspace to use, if not provided
-            personal workspace will be used.
-        cs_endpoint (str, optional): endpoint to use for the webserver, this will connect
-          to the custom ports functionality of the instance. Defaults to ``server``,
+      i (str): name or `project_id` of the instance
     """
     if not i:
-      raise ValueError("Instance id must be provided, try --i='8h57f9'")
+      raise ValueError("Instance id must be provided, try --i='1023'")
     i = str(i)
 
     # simply add useful keys to the instance
     self.project_id: str = None
     self.project_name: str = None
-    self.workspace_id: str = workspace_id
+    self.workspace_id: str = workspace_id or secret(AuthConfig.workspace_id)
     self.size_used: float = None
     self.size: float = None
     self.state: str = None
@@ -115,25 +110,31 @@ class Instance():
 
     # create a new subway for webserver
     sess = Session()
-    sess.headers.update({"Authorization": f"Bearer {secret.get('access_token')}"})
+    sess.headers.update({"Authorization": f"Bearer {secret('access_token')}"})
     stub_ws_instance = create_webserver_subway("v1", sess)
-    if workspace_id == None:
+    if self.workspace_id == None:
       stub_projects = stub_ws_instance.user.projects
     else:
-      stub_projects = stub_ws_instance.workspace.u(workspace_id).projects
+      stub_projects = stub_ws_instance.workspace.u(self.workspace_id).instances
 
     # filter and get the data
     project_details = stub_projects()["project_details"]
+    # print(project_details)
     if i not in project_details:
-      by_name = list(filter(lambda x: x['project_name'] == i, list(project_details.values())))
+      by_name = list(filter(lambda x: x[1]['project_name'] == i, list(project_details.items())))
       if len(by_name) == 0:
         raise ValueError(f"Instance '{i}' not found")
       elif len(by_name) > 1:
         raise ValueError(f"Multiple instances with name '{i}' found")
       data = by_name[0]
+      project_id = data[0]
+      data = data[1]
     else:
       data = project_details[i]
+      project_id = i
+    data["project_id"] = project_id
     logger.info(f"Found instance '{data['project_name']}' ({data['project_id']})")
+    # print(data)
     for x in self.useful_keys:
       setattr(self, x, data[x])
     
@@ -160,24 +161,88 @@ class Instance():
   ################################################################################
 
   @classmethod
-  def new(cls, project_name: str, workspace_id: str = None, storage_limit: int = 25, project_type = "blank") -> 'Instance':
-    if workspace_id == None:
-      stub_all_projects = nbox_ws_v1.user.projects
-    else:
-      stub_all_projects = nbox_ws_v1.workspace.u(workspace_id).projects
-    out = stub_all_projects(
-      _method = "post",
-      project_name = project_name,
-      storage_limit = storage_limit,
-      project_type = project_type,
-      github_branch = "",
-      github_link = "",
-      template_id = 0,
-      clone_id = 0
-  )
+  def new(
+    cls,
+    project_name: str,
+    storage_limit: int = 25,
+    project_type: str = "blank",
+    github_branch: str = "",
+    github_link: str = "",
+    template_id: int = 0,
+    clone_id: int = 0,
+    *,
+    workspace_id: str = "",
+  ) -> 'Instance':
+    """Create a new NBX-Build instance.
+    
+    Args:
+      project_name (str): Name of the instance
+      storage_limit (int, optional): Storage limit in GB. Defaults to 25.
+      project_type (str, optional): Type of the instance. Defaults to "blank".
+      github_branch (str, optional): Branch of the github repo. Defaults to "".
+      github_link (str, optional): Link to the github repo. Defaults to "".
+      template_id (int, optional): ID of the template to use. Defaults to 0.
+      clone_id (int, optional): ID of the instance to clone. Defaults to 0.
 
+    Returns:
+      Instance: The newly created instance.
+    """
+    workspace_id = workspace_id or secret(AuthConfig.workspace_id)
+    stub_all_projects = nbox_ws_v1.projects
+
+    kwargs_dict = {
+      "project_name": project_name,
+      "storage_limit": storage_limit,
+      "project_type": project_type,
+    }
+    if github_branch:
+      kwargs_dict["github_branch"] = github_branch
+    if github_link:
+      kwargs_dict["github_link"] = github_link
+    if template_id:
+      kwargs_dict["template_id"] = template_id
+    if clone_id:
+      kwargs_dict["clone_id"] = clone_id
+
+    out = stub_all_projects(_method = "post", **kwargs_dict)
     return out
 
+  ################################################################################
+  """
+  # Utility Methods
+
+  Functions are responsible for doing things that otherwise the user would struggle with.
+  """
+  ################################################################################
+
+  def get_subway(self, subdomain) -> Sub30:
+    """Get a Subway object for the instance.
+
+    Args:
+      subdomain (str): The subdomain to connect to
+    
+    Returns:
+      Subway: The Subway object.
+    """
+    self._unopened_error()
+
+    build = "build"
+    if "app.c." in secret("nbx_url"):
+      build = "build.c"
+    elif "app.rc" in secret("nbx_url"):
+      build = "build.rc"
+    url = f"https://{subdomain}-{self.open_data['url']}.{build}.nimblebox.ai/",
+    session = Session(
+      headers = {
+        "NBX-TOKEN": self.open_data["token"],
+        "X-NBX-USERNAME": secret("username"),
+      }
+    )
+    r = session.get(url + "openapi.json")
+    data = r.json()
+    sub = Sub30(url, data, session)
+    return sub
+
   ################################################################################
   """
   # State Methods
@@ -190,7 +255,7 @@ class Instance():
     """Check if the instance is running.
 
     Returns:
-        bool: True if the instance is running, False otherwise.
+      bool: True if the instance is running, False otherwise.
     """
     return self.state == "RUNNING"
 
@@ -241,7 +306,8 @@ class Instance():
       base_domain = launch_data['base_domain']
       self.open_data = {
         "url": f"{base_domain}",
-        "token": self.stub_ws_instance._session.cookies.get_dict()[f"instance_token_{base_domain}"]
+        "token": self.stub_ws_instance._session.cookies.get_dict()[f"instance_id_{base_domain}"],
+        "launch_url": launch_data['launch_url']
       }
       self.__opened = True
 
@@ -261,12 +327,12 @@ class Instance():
     Actual start is implemented in `_start` method, this combines other things
 
     Args:
-        cpu (int, optional): CPU count should be one of ``[2, 4, 8]``
-        gpu (str, optional): GPU name should be one of ``["t5", "p100", "v100", "k80"]``
-        gpu_count (int, optional): When zero, cpu-only instance is started
-        auto_shutdown (int, optional): No autoshutdown if zero, defaults to 6.
-        dedicated_hw (bool, optional): If not spot/pre-emptible like machines used
-        zone (str, optional): GCP cloud regions, defaults to "asia-south-1".
+      cpu (int, optional): CPU count should be one of ``[2, 4, 8]``
+      gpu (str, optional): GPU name should be one of ``["t5", "p100", "v100", "k80"]``
+      gpu_count (int, optional): When zero, cpu-only instance is started
+      auto_shutdown (int, optional): No autoshutdown if zero, defaults to 6.
+      dedicated_hw (bool, optional): If not spot/pre-emptible like machines used
+      zone (str, optional): GCP cloud regions, defaults to "asia-south-1".
     """
     
 
@@ -285,7 +351,7 @@ class Instance():
       self._open()
 
   def stop(self):
-    """Stop Instance"""
+    """Stop the Instance"""
     if self.state == "STOPPED":
       logger.info(f"Instance {self.project_name} ({self.project_id}) is already stopped")
       return
@@ -349,7 +415,7 @@ class Instance():
 
     conman = ConnectionManager(
       file_logger = file_logger,
-      user = secret.get("username"), 
+      user = secret("username"), 
       subdomain = self.open_data.get("url"),
       auth = self.open_data.get("token"),
     )
@@ -383,14 +449,17 @@ class Instance():
     connection.quit()
     return result.strip()
 
-  def ls(self, path: str, *, port: int = 6174):
+  def ls(self, path: str, *, port: int = 6174) -> str:
     """
     List files in a directory relative to '/home/ubuntu/project'
 
     ```
     [nbox API] - nbox.Instance(...).ls("/")
-    [nbox CLI] - python3 -m nbox build ... ls ./
+    [nbox CLI] - nbx build ... ls ./
     ```
+
+    Args:
+      path (str): Path to list files in all paths will be built relative to "/home/ubuntu/project/" folder.
     """
     self._unopened_error()
     if path.startswith("nbx://"):
@@ -410,10 +479,13 @@ class Instance():
 
   def mv(self, src: str, dst: str, force: bool = False, *, port: int = 6174):
     """
-    Move files to and fro NBX-Build.
+    Move files to and fro NBX-Build. Use 'nbx://' as prefix for Instance, all files will be placed relative to
+    '/home/ubuntu/project/' folder.
 
-    Use 'nbx://' as prefix for Instance, all files will be placed relative to
-    "/home/ubuntu/project/" folder.
+    Args:
+      src (str): Source file or folder to move
+      dst (str): Destination file or folder to move to
+      force (bool): If True, will override the destination file if it already exists
     """
     self._unopened_error()
 
@@ -466,9 +538,12 @@ class Instance():
     result = self.__run_command(comm, port)
     return result
 
-  def rm(self, file: str, *, port: int = 6174):
+  def rm(self, file: str, *, port: int = 6174) -> str:
     """
     Remove file from NBX-Build.
+
+    Args:
+      file (str): File to remove
     """
     self._unopened_error()
 
@@ -485,10 +560,12 @@ class Instance():
     result = self.__run_command(comm, port)
     return result
 
-  def remote(self, x: str, *, port: int = 6174):
+  def remote(self, x: str, *, port: int = 6174) -> str:
     """
-    Run any command using SSH, in the underlying system it will setup a SSH connection and execute
-    any command.
+    Run any command using SSH, in the underlying system it will setup a SSH connection and execute any command.
+
+    Args:
+      x (str): Command to run
     """
     self._unopened_error()
 
@@ -518,5 +595,5 @@ class Instance():
     any IDE with checker is dead easy, however the performace guarantee is premium given high
     costs. Thus the logic to parsing these will have to be written as a seperate module.
     """
-    raise NotImplementedError("Core talking method, not implemented yet")
     # self._unopened_error()
+    raise NotImplementedError("Core talking method, not implemented yet")
