@@ -1,14 +1,16 @@
 """
 Jupyter Notebooks are user by millions of developers world wide, so we said why not just
 create an operator that runs the jupyter notebook and creates a python file out of it.
-Catch: Currently we do not support notebooks that have a ``!`` for shell commands, please
-replace them with ``subprocess.call(shelex.split(<com>))``
+Catch: Currently we do not support notebooks that have a `!` for shell commands, please
+replace them with `subprocess.call(shelex.split(<com>))`
 
-.. code-block:: python
+Example:
 
-  from nbox.lib.notebook import NotebookRunner
-  op = NotebookRunner("./sample.ipynb")
-  op() # call and it will do the rest
+```
+from nbox.lib.notebook import NotebookRunner
+op = NotebookRunner("./sample.ipynb")
+op() # call and it will do the rest
+```
 """
 
 import sys
@@ -26,7 +28,10 @@ def get_py_builtin_all():
   # a cheap 99% solution that works for 99% of people
   builtin_path = U.join(U.folder(U.folder(__file__)), "assets/builtin")
   with open(builtin_path, "r") as f:
-    items = set(f.read().splitlines())
+    items = []
+    for l in f:
+      if not l.startswith("#"):
+        items.append(l.strip())
   return items
 
 def get_package_version(name = 'tensorflow'):
