@@ -2,7 +2,6 @@
 NimbleBox LMAO is our general purpose observability tool for any kind of computation you might have.
 """
 
-# f"{lmao.project_id}/{lmao.run.experiment_id}/model.pkl"
 # drift detection and all
 # run.log_dataset(
 #     dataset_name='train',
@@ -17,92 +16,47 @@ NimbleBox LMAO is our general purpose observability tool for any kind of computa
 #     actuals=y_test,
 # )
 
-from functools import lru_cache
 import os
 import re
 import sys
-import time
 import shlex
 import zipfile
-import threading
-from queue import Queue, Empty
 from git import Repo
-from json import dumps, load
+from json import dumps, loads
+from functools import lru_cache
 from requests import Session
-from typing import Dict, Any, List, Optional, Union, Tuple
-from subprocess import Popen, PIPE
+from typing import Dict, Any, List, Optional, Union
+from subprocess import Popen
 from google.protobuf.field_mask_pb2 import FieldMask
 
-try:
-  import starlette
-  from starlette.requests import Request
-  from starlette.responses import Response
-  from starlette.middleware.base import BaseHTTPMiddleware, RequestResponseEndpoint
-  from starlette.routing import Match
-  from starlette.status import HTTP_500_INTERNAL_SERVER_ERROR
-  from starlette.types import ASGIApp
-except ImportError:
-  starlette = None
-
-
 import nbox.utils as U
 from nbox import Instance
 from nbox.utils import logger, SimplerTimes
-from nbox.auth import secret, ConfigString
+from nbox.auth import secret, AuthConfig, auth_info_pb
 from nbox.nbxlib.tracer import Tracer
-from nbox.relics import RelicsNBX
-from nbox.jobs import Job, Resource
-from nbox.init import nbox_grpc_stub
+from nbox.relics import Relics
+from nbox.jobs import Job, upload_job_folder
+from nbox.hyperloop.common.common_pb2 import Resource
+from nbox.init import nbox_grpc_stub, nbox_ws_v1
 from nbox.messages import message_to_dict
-from nbox.hyperloop.nbox_ws_pb2 import UpdateJobRequest
-from nbox.hyperloop.job_pb2 import Job as JobProto
+from nbox.hyperloop.jobs.nbox_ws_pb2 import UpdateJobRequest, JobRequest
+from nbox.hyperloop.jobs.job_pb2 import Job as JobProto
 
 # all the sublime -> hyperloop stuff
-from nbox.sublime.lmao_client import LMAO_Stub # main stub class
-from nbox.sublime.lmao_client import (
-  Record, File, FileList, AgentDetails, RunLog, Run, InitRunRequest, ListProjectsRequest, RelicFile
+from nbox.sublime.lmao_rpc_client import (
+  LMAO_Stub, # main stub class
+  Record, File, FileList, AgentDetails, RunLog,
+  Run, InitRunRequest, ListProjectsRequest,
 )
-from nbox.sublime.lmao_client import (
-  Serving, LogBuffer, ServingHTTPLog
-)
-
-
 from nbox.observability.system import SystemMetricsLogger
 
-DEBUG_LOG_EVERY = 100
-INFO_LOG_EVERY = 100
 
 """
 functional components of LMAO
 """
 
-@lru_cache()
-def get_lmao_stub(username: str, workspace_id: str):
-  # prepare the URL
-  id_or_name = f"monitoring-{workspace_id}"
-  logger.info(f"Instance id_or_name: {id_or_name}")
-  logger.debug(f"workspace_id: {workspace_id}")
-  instance = Instance(id_or_name, workspace_id = workspace_id)
-  try:
-    open_data = instance.open_data
-  except AttributeError:
-    raise Exception(f"Is instance '{instance.project_id}' running?")
-  build = "build"
-  if "app.c." in secret.get("nbx_url"):
-    build = "build.c"
-  url = f"https://server-{open_data['url']}.{build}.nimblebox.ai/"
-  logger.debug(f"URL: {url}")
-
-  # create a session with the auth header
-  _session = Session()
-  _session.headers.update({
-    "NBX-TOKEN": open_data["token"],
-    "X-NBX-USERNAME": username,
-  })
-
-  # define the stub
-  # self.lmao = LMAO_Stub(url = "http://127.0.0.1:8080", session = Session()) # debug
-  lmao_stub = LMAO_Stub(url = url, session = _session)
+def get_lmao_stub() -> LMAO_Stub:
+  lmao_stub = LMAO_Stub(url = secret(AuthConfig.url) + "/monitoring", session = nbox_ws_v1._session)
   return lmao_stub
 
 def get_record(k: str, v: Union[int, float, str]) -> Record:
@@ -124,7 +78,7 @@ def get_record(k: str, v: Union[int, float, str]) -> Record:
   return record
 
 def get_git_details(folder):
-  """If there is a .git folder in the folder, return some details for that."""
+  """If there is a `.git` folder in the folder, return some details for that."""
   repo = Repo(folder)
 
   # check for any unstaged files
@@ -170,37 +124,26 @@ Client library that the user will use to interact with the LMAO server.
 """
 
 class _lmaoConfig:
+  # _lmaoConfig.kv contains all the objects that the class LMAO needs to work correctly, however
+  # we will also need to take care of the things we want to show 
   kv = {}
   def set(
-    workspace_id: str = "",
-    project_id: Optional[str] = "",
-    project_name: Optional[str] = "",
-    experiment_id: Optional[str] = "",
-    metadata: Dict[str, Any] = {},
-    save_to_relic: bool = False,
-    enable_system_monitoring: bool = False,
-    args = (),
-    kwargs = {},
+    project_name: str,
+    project_id: str,
+    experiment_id: str,
+    save_to_relic: bool,
+    enable_system_monitoring: bool,
+    store_git_details: bool,
   ) -> None:
     _lmaoConfig.kv = {
-      "workspace_id": workspace_id,
-      "project_id": project_id,
       "project_name": project_name,
+      "project_id": project_id,
       "experiment_id": experiment_id,
-      "metadata": metadata,
       "save_to_relic": save_to_relic,
       "enable_system_monitoring": enable_system_monitoring,
-      "args": args,
-      "kwargs": kwargs,
+      "store_git_details": store_git_details,
     }
 
-  def clear() -> None:
-    _lmaoConfig.kv = {}
-
-  def json() -> str:
-    # return dumps({k:v for k,v in _lmaoConfig.kv.items() if k not in ["args", "kwargs"]})
-    return dumps(_lmaoConfig.kv)
-
 
 class Lmao():
   def __init__(
@@ -211,13 +154,23 @@ class Lmao():
     metadata: Dict[str, Any] = {},
     save_to_relic: bool = False,
     enable_system_monitoring: bool = False,
-    workspace_id: str = "",
+    store_git_details: bool = True,
   ) -> None:
-    """``Lmao`` is the client library for using NimbleBox Monitoring. It talks to your monitoring instance running on your build
-    and stores the information in the ``project_name`` or ``project_id``. This object inherently doesn't care what you are actually
+    """`Lmao` is the client library for using NimbleBox Monitoring. It talks to your monitoring instance running on your build
+    and stores the information in the `project_name` or `project_id`. This object inherently doesn't care what you are actually
     logging and rather concerns itself with ensuring storage.
-    
-    All arguments are optional, if the _lmaoConfig is set"""
+
+    **Note**: All arguments are optional, if the `_lmaoConfig.kv` is set.
+
+    Args:
+      project_name (str, optional): The name of the project. Defaults to "".
+      project_id (str, optional): The id of the project. Defaults to "".
+      experiment_id (str, optional): The id of the experiment. Defaults to "".
+      metadata (Dict[str, Any], optional): Any metadata that you want to store. Defaults to {}.
+      save_to_relic (bool, optional): Whether to save the data to the relic. Defaults to False.
+      enable_system_monitoring (bool, optional): Whether to enable system monitoring. Defaults to False.
+      store_git_details (bool, optional): Whether to store git details. Defaults to True.
+    """
 
     self.config = _lmaoConfig.kv
 
@@ -226,26 +179,23 @@ class Lmao():
       project_name = _lmaoConfig.kv["project_name"]
       project_id = _lmaoConfig.kv["project_id"]
       experiment_id = _lmaoConfig.kv["experiment_id"]
-      metadata = _lmaoConfig.kv["metadata"]
       save_to_relic = _lmaoConfig.kv["save_to_relic"]
       enable_system_monitoring = _lmaoConfig.kv["enable_system_monitoring"]
-      workspace_id = _lmaoConfig.kv["workspace_id"]
+      store_git_details = _lmaoConfig.kv["store_git_details"]
 
     self.project_name = project_name
     self.project_id = project_id
     self.experiment_id = experiment_id
-    self.metadata = metadata
     self.save_to_relic = save_to_relic
     self.enable_system_monitoring = enable_system_monitoring
-    self.workspace_id = workspace_id
-    
+    self.store_git_details = store_git_details
+    self.workspace_id = secret(AuthConfig.workspace_id)
+
     # now set the supporting keys
     self.nbx_job_folder = U.env.NBOX_JOB_FOLDER("")
     self._total_logged_elements = 0 # this variable keeps track for logging
     self.completed = False
-    self.tracer = None
-    self.username = secret.get("username")
-    self.relic: RelicsNBX = None
+    self.relic: Relics = None
     self.system_monitoring: SystemMetricsLogger = None
     self._nbx_run_id = None
     self._nbx_job_id = None
@@ -258,51 +208,40 @@ class Lmao():
       self.system_monitoring.stop()
 
   def _get_name_id(self, project_id: str = None, project_name: str = None):
-    all_projects = self.lmao.list_projects(
-      _ListProjectsRequest = ListProjectsRequest(
+    matching_projects = self.lmao.list_projects(
+      ListProjectsRequest(
         workspace_id = self.workspace_id,
         project_id_or_name = project_id if project_id else project_name,
       )
     )
-    if not all_projects.projects:
+
+    # case for one matching project
+    if len(matching_projects.projects) == 1:
+      project_name = matching_projects.projects[0].project_name
+      project_id = matching_projects.projects[0].project_id
+    elif len(matching_projects.projects) > 1:
+      if project_id:
+        logger.error(f"Multiple entries for {project_id} found, something went wrong from our side.")
+        raise Exception(f"Database duplicate entry for project_id: {project_id}")
+      else:
+        logger.error(f"Project '{project_name}' found multiple times, please use project_id instead.")
+        raise Exception("Ambiguous project name, please use project_id instead.")
+    else:
+      # case for no matching projects returned
       # means we have to initialise this project first and then we will pull the details after the init_run
       if project_id:
         raise Exception(f"Project with id {project_id} not found, please create a new one with project_name")
       logger.info(f"Project '{project_name}' not found, will create a new one.")
-    
-    if project_id:
-      common = list(filter(lambda x: x.project_id == project_name, all_projects.projects))
-      if len(common) == 1:
-        project_name = common[0].project_name
-        project_id = common[0].project_id
-      elif len(common) > 1:
-        logger.error(f"Multiple entries for {project_id} found, something went wrong from our side.")
-        raise Exception(f"Database duplicate entry for project_id: {project_id}")
-      else:
-        raise Exception(f"Project with id {project_id} not found, please create a new one with project_name")
-    else:
-      common = list(filter(lambda x: x.project_name == project_name, all_projects.projects))
-      if len(common) == 1:
-        project_name = common[0].project_name
-        project_id = common[0].project_id
-      elif len(common) > 1:
-        logger.error(f"Project '{project_name}' found multiple times, please use project_id instead.")
-        raise Exception("Ambiguous project name, please use project_id instead.")
-      else:
-        raise Exception(f"Project with name {project_name} not found, please create a new one with project_name")
+
     return project_id, project_name
 
   def _init(self, project_name, project_id, config: Dict[str, Any] = {}):
     # create a tracer object that will load all the information
-    self.tracer = Tracer(start_heartbeat=False)
-    self._nbx_run_id = self.tracer.run_id
-    self._nbx_job_id = self.tracer.job_id
-    if self._nbx_run_id is not None:
-      # update the username since jobs pod does not contain that information
-      secret.put("username", self.tracer.job_proto.auth_info.username)
-      self.username = secret.get("username")
-
-    self.lmao = get_lmao_stub(self.username, self.workspace_id)
+    tracer = Tracer(start_heartbeat = False)
+    self._nbx_run_id = tracer.run_id
+    self._nbx_job_id = tracer.job_id
+    del tracer
+    self.lmao = get_lmao_stub()
 
     # do a quick lookup and see if the project exists, if not, create it
     if self.config:
@@ -313,10 +252,12 @@ class Lmao():
         project_id, project_name = self._get_name_id(project_id=project_id)
         if not project_id:
           raise Exception(f"Project with id {project_id} not found, please create a new one with project_name")
-      else:
+      elif project_name:
         project_id, project_name = self._get_name_id(project_name=project_name)
         if not project_id:
           logger.info(f"Project '{project_name}' not found, create one from dashboard.")
+      else:
+        raise Exception("provide either `project_id` or `project_name`")
 
     # this is the config value that is used to store data on the plaform, user cannot be allowed to have
     # like a full access to config values
@@ -327,58 +268,69 @@ class Lmao():
     # check if the current folder from where this code is being executed has a .git folder
     # NOTE: in case of NBX-Jobs the current folder ("./") is expected to contain git by default
     log_config["git"] = None
-    if os.path.exists(".git"):
+    if os.path.exists(".git") and self.store_git_details:
       log_config["git"] = get_git_details("./")
 
     # continue as before
     self._agent_details = AgentDetails(
       type = AgentDetails.NBX.JOB,
-      nbx_job_id = self._nbx_job_id or "jj_guvernr",
-      nbx_run_id = self._nbx_run_id or "fake_run",
+      nbx_job_id = self._nbx_job_id or "",
+      nbx_run_id = self._nbx_run_id or "",
     )
 
-    run_details = self.lmao.get_run_details(Run(
-      experiment_id = _lmaoConfig.kv["experiment_id"],
-    ))
-    if run_details.experiment_id:
-      # means that this run already exists so we need to make an update call
-      ack = self.lmao.update_run_status(Run(
-        experiment_id = run_details.experiment_id,
-        agent = self._agent_details,
-      ))
-      if not ack.success:
-        raise Exception(f"Failed to update run status!")
+    if self.experiment_id:
+      action = "Updated"
+      run_details = self.lmao.get_run_details(
+        Run(
+          workspace_id = self.workspace_id,
+          project_id=project_id,
+          experiment_id = self.experiment_id,
+        )
+      )
+      if not run_details:
+        # TODO: Make a custom exception of this
+        raise Exception("Server Side exception has occurred, Check the log for details")
+      if run_details.experiment_id:
+        # means that this run already exists so we need to make an update call
+        ack = self.lmao.update_run_status(
+          Run(
+            workspace_id = self.workspace_id,
+            project_id = project_id,
+            experiment_id = run_details.experiment_id,
+            agent = self._agent_details,
+            update_keys = ["agent"],
+          )
+        )
+        if not ack.success:
+          raise Exception(f"Failed to update run status! {ack.message}")
     else:
-      # check if there is a lmao run existing for this project_id
+      action = "Created"
       run_details = self.lmao.init_run(
-        _InitRunRequest = InitRunRequest(
+        InitRunRequest(
+          workspace_id = self.workspace_id,
           agent_details=self._agent_details,
-          created_at = SimplerTimes.get_now_i64(),
           project_name = project_name,
           project_id = project_id,
           config = dumps(log_config),
         )
       )
 
-    if not run_details:
-      # TODO: Make a custom exception of this
-      raise Exception("Server Side exception has occurred, Check the log for details")
-
     self.project_name = project_name
-    self.project_id = project_id
-    self.metadata = config
-
+    self.project_id = run_details.project_id
     self.run = run_details
-    logger.info(f"Created a new LMAO run")
-    logger.info(f" project: {self.project_name} ({self.project_id})")
-    logger.info(f"      id: {self.run.experiment_id}")
-    logger.info(f"    link: https://app.nimblebox.ai/workspace/{self.workspace_id}/monitoring/{self.project_id}/{self.run.experiment_id}")
+
+    logger.info(
+      f"{action} LMAO run\n"
+      f" project: {self.project_name} ({self.project_id})\n"
+      f"      id: {self.run.experiment_id}\n"
+      f"    link: https://app.nimblebox.ai/workspace/{self.workspace_id}/monitoring/{self.project_id}/{self.run.experiment_id}\n"
+    )
 
     # now initialize the relic
     if self.save_to_relic:
       # The relic will be the project id
-      self.relic = RelicsNBX("experiments", self.workspace_id, create = True)
-      logger.info(f"Will store everything in folder: {self.experiment_prefix}")
+      self.relic = Relics(LMAO_RELIC_NAME, self.workspace_id, create = True)
+      logger.info(f"Will store everything in relic '{LMAO_RELIC_NAME}' folder: {self.experiment_prefix}")
 
     # system metrics monitoring, by default is enabled optionally turn it off
     if self.enable_system_monitoring:
@@ -387,15 +339,14 @@ class Lmao():
 
   @property
   def experiment_prefix(self):
-    prefix = f"{self.project_name}/{self.run.experiment_id}/"
-    return prefix
+    return f"{self.project_name}/{self.run.experiment_id}/"
 
   """The functions below are the ones supposed to be used."""
 
   @lru_cache(maxsize=1)
   def get_relic(self):
     """Get the underlying Relic for more advanced usage patterns."""
-    return RelicsNBX("experiments", self.workspace_id, create = True, prefix = f"{self.project_name}/{self.run.experiment_id}")
+    return Relics(LMAO_RELIC_NAME, self.workspace_id, create = True, prefix = f"{self.experiment_prefix}")
 
   def log(self, y: Dict[str, Union[int, float, str]], step = None, *, log_type: str = RunLog.LogType.USER):
     """Log a single level dictionary to the platform at any given step. This function does not really care about the
@@ -403,25 +354,24 @@ class Lmao():
     if self.completed:
       raise Exception("Run already completed, cannot log more data!")
 
-    # if self._total_logged_elements % DEBUG_LOG_EVERY == 0:
-    #   logger.debug(f"Logging: {y.keys()} | {self._total_logged_elements}")
-    # if self._total_logged_elements % INFO_LOG_EVERY == 0:
-    #   logger.info(f"Logging: {y.keys()} | {self._total_logged_elements}")
-
     step = step if step is not None else SimplerTimes.get_now_i64()
     if step < 0:
       raise Exception("Step must be <= 0")
-    run_log = RunLog(experiment_id = self.run.experiment_id, log_type=log_type)
+    run_log = RunLog(
+      workspace_id = self.workspace_id,
+      project_id=self.project_id,
+      experiment_id = self.run.experiment_id,
+      log_type=log_type
+    )
     for k,v in y.items():
+      # TODO:@yashbonde replace Record with RecordColumn
       record = get_record(k, v)
       record.step = step
       run_log.data.append(record)
 
-    ack = self.lmao.on_log(_RunLog = run_log)
+    ack = self.lmao.on_log(run_log)
     if not ack.success:
-      logger.error("  >> Server Error")
-      for l in ack.message.splitlines():
-        logger.error("  " + l)
+      logger.error(f"  >> Server Error\n{ack.message}")
       raise Exception("Server Error")
 
     self._total_logged_elements += 1
@@ -431,12 +381,10 @@ class Lmao():
     Register a file save. User should be aware of some structures that we follow for standardizing the data.
     All the experiments are going to be tracked under the following pattern:
 
-    #. ``relic_name`` is going to be the experiment ID, so any changes to the name will not affect relic storage
-    #. ``{experiment_id}(_{job_id}@{experiment_id})`` is the name of the folder which contains all the artifacts in the experiment.
+    If `save_to_relics` is not set, this function is a no-op.
 
-    If relics is not enabled, this function will simply log to the LMAO DB.
-
-    dk.save_file("foo.t", "/bar/", "baz.t", "/bar/roo/")
+    Args:
+      files: The list of files to save. This can be a list of files or a list of folders. If a folder is passed, all the files in the folder will be uploaded.
     """
     logger.info(f"Saving files: {files}")
 
@@ -446,10 +394,12 @@ class Lmao():
       if os.path.isfile(folder_or_file):
         all_files.append(folder_or_file)
       elif os.path.isdir(folder_or_file):
-        all_files.extend(U.get_files_in_folder(folder_or_file, self.username, self.workspace_id))
+        all_files.extend(U.get_files_in_folder(folder_or_file))
       else:
         raise Exception(f"File or Folder not found: {folder_or_file}")
 
+    # TODO: @yashbonde log the files in the LMAO DB for sanity, currently this is a no-op, keeping it here so one day
+    # when we add something cool we can use this
     logger.debug(f"Storing {len(all_files)} files")
     if self.save_to_relic:
       relic = self.get_relic()
@@ -457,11 +407,6 @@ class Lmao():
       for f in all_files:
         relic.put(f)
 
-    # log the files in the LMAO DB for sanity
-    fl = FileList(experiment_id = self.run.experiment_id)
-    fl.files.extend([File(relic_file = RelicFile(name = x)) for x in all_files])
-    self.lmao.on_save(_FileList = fl)
-
   def end(self):
     """End the run to declare it complete. This is more of a convinience function than anything else. For example when you
     are monitoring a live API you may never know when the experiment is complete. This function locks the experiment name
@@ -471,7 +416,7 @@ class Lmao():
       return None
 
     logger.info("Ending run")
-    ack = self.lmao.on_train_end(_Run = Run(experiment_id=self.run.experiment_id,))
+    ack = self.lmao.on_train_end(self.run)
     if not ack.success:
       logger.error("  >> Server Error")
       for l in ack.message.splitlines():
@@ -481,121 +426,76 @@ class Lmao():
     if self.enable_system_monitoring:
       self.system_monitoring.stop()
 
-"""
-Code responsible for Monitoring of live servings
-"""
-
-
-def post_buffer(lmao: LMAO_Stub, queue: Queue, bar: threading.Barrier):
-  while True:
-    bar.wait() # rate_limit
-    buff = LogBuffer()
-    while True:
-      try:
-        item = queue.get()
-        buff.logs.append(item)
-      except Empty:
-        break
-    lmao.on_livelog(buff)
-
-
-class LmaoASGIMiddleware():
-
-  def __init__(self, app, workspace_id: str = "") -> None:
-    if starlette is None:
-      raise ValueError("Starlette is not installed. pip install -U nbox\[serving\]")
-    super().__init__(app)
-
-    # rate limiter
-    self._bar = threading.Barrier(2)
-    def _rate_limiter(s = 1.0):
-      while True:
-        self._bar.wait()
-        time.sleep(s)
-    rl = threading.Thread(target=_rate_limiter, daemon=True)
-
-    # create connection and handshake with the lmao server
-    self.workspace_id = workspace_id
-
-    # create a tracer object that will load all the information
-    self.tracer = Tracer(start_heartbeat=False)
-    self._nbx_run_id = self.tracer.run_id
-    self._nbx_job_id = self.tracer.job_id
-    if self._nbx_run_id is not None:
-      # update the username you have
-      secret.put("username", self.tracer.job_proto.auth_info.username)
-      self.username = secret.get("username")
-    self.lmao = LMAO_Stub(self.username, self.workspace_id)
-
-    serving = self.lmao.init_serving(
-      InitRunRequest(
-        agent_details = AgentDetails(
-          type = AgentDetails.NBX.SERVING,
-          nbx_serving_id = self._nbx_job_id or "jj_guvernr",
-          nbx_run_id = self._nbx_run_id or "fake_deploy",
-        ),
-        created_at = SimplerTimes.get_now_i64(),
-        config = dumps({
-          "my_config": "config"
-        })
-      )
-    )
-    self.serving = serving
-    self.buffer = Queue(maxsize = 2 << 14)
-    self.logger_thread = threading.Thread(target = post_buffer, args = (self.lmao, self.buffer, self._bar), daemon = True)
-    rl.start()
-    self.logger_thread.start()
-
-  async def dispatch(self, request, call_next):
-    method = request.method
-    path_template, is_handled_path = self.get_path_template(request)
-    if not is_handled_path:
-      return await call_next(request)
-
-    before_time = time.perf_counter()
-    try:
-      response = await call_next(request)
-    except BaseException as e:
-      status_code = HTTP_500_INTERNAL_SERVER_ERROR
-      raise e from None
-    else:
-      status_code = response.status_code
-      after_time = time.perf_counter()
-
-    path = path_template.strip("/")
-    if path:
-      _log = ServingHTTPLog(
-        path = path,
-        method = method,
-        status_code = status_code,
-        latency_ms = 1e3 * (after_time-before_time),
-        timestamp = SimplerTimes.get_now_pb(),
-      )
-      self.buffer.put_nowait(_log)
-
-    return response
-
-  @staticmethod
-  def get_path_template(request) -> Tuple[str, bool]:
-    for route in request.app.routes:
-      match, child_scope = route.matches(request.scope)
-      if match == Match.FULL:
-        return route.path, True
-    return request.url.path, False
-
-
 """
 For some Experiences you want to have CLI control so this class manages that
 
+```
 lmao upload fp:fn project_name_or_id
 lmao trigger project_name_or_id
 lmao open project_name_or_id
+```
 """
 
+class ExperimentConfig:
+  def __init__(
+    self,
+    run_kwargs: Dict[str, Any],
+    git: Dict[str, Any],
+    resource: Resource,
+    cli_comm: str,
+    save_to_relic: bool = True,
+    enable_system_monitoring: bool = False,
+  ):
+    """In an ideal world this would be a protobuf message, but we are not there yet. This contains all the
+    things that are stored in the DB and FE can use this to render elements in the UI.
+
+    Args:
+      run_kwargs (Dict[str, Any]): All the arguments that the user passed to the `nbx lmao run ...` CLI
+      git (Dict[str, Any]): details for the git repo
+      resource (Resource): Resource pb object that contains the details of the resource
+      cli_comm (str): The CLI command that was used to run the experiment, user can use this to reproduce the experiment
+      save_to_relic (bool): If the user wants to save the experiment to the relic by default
+      enable_system_monitoring (bool): If the user wants to enable system monitoring
+    """
+    self.run_kwargs = run_kwargs
+    self.git = git
+    self.resource = resource
+    self.cli_comm = cli_comm
+    self.save_to_relic = save_to_relic
+    self.enable_system_monitoring = enable_system_monitoring
+
+  def to_dict(self):
+    return {
+      "run_kwargs": self.run_kwargs,
+      "git": self.git,
+      "resource": message_to_dict(self.resource),
+      "cli_comm": self.cli_comm,
+      "save_to_relic": self.save_to_relic,
+      "enable_system_monitoring": self.enable_system_monitoring,
+    }
+
+  def to_json(self):
+    return dumps(self.to_dict())
+
+  @classmethod
+  def from_json(cls, json_str):
+    d = loads(json_str)
+    print(d["resource"])
+    d["resource"] = Resource(
+      cpu = str(d["resource"]["cpu"]),
+      memory = str(d["resource"]["memory"]),
+      disk_size = str(d["resource"]["disk_size"]),
+      gpu = str(d["resource"]["gpu"]),
+      gpu_count = str(d["resource"]["gpu_count"]),
+      timeout = int(d["resource"]["timeout"]),
+      max_retries = int(d["resource"]["max_retries"]),
+    )
+    return cls(**d)
+
+
 @lru_cache()
 def get_project_name_id(project_name_or_id: str, workspace_id: str):
-  username = secret.get("username")
-  lmao_stub = get_lmao_stub(username, workspace_id)
+  lmao_stub = get_lmao_stub()
   out = lmao_stub.list_projects(ListProjectsRequest(workspace_id=workspace_id, project_id_or_name=project_name_or_id))
   if not out.projects:
     logger.error(f"Project: {project_name_or_id} not found")
@@ -613,11 +513,13 @@ def get_project_name_id(project_name_or_id: str, workspace_id: str):
 
 class LmaoCLI:
   # This class will only manage the things that need to talk to the DB and everything else is offloaded
-  def upload(
+  def upload(self, **kwargs):
+    raise ValueError("`nbx lmao upload` is not changed to `nbx lmao run`")
+
+  def run(
     self,
     init_path: str,
     project_name_or_id: str,
-    workspace_id: str = "",
     trigger: bool = False,
 
     # all the arguments for the git thing
@@ -646,10 +548,8 @@ class LmaoCLI:
     """Upload and register a new run for a NBX-LMAO project.
 
     Args:
-      init_path (str): This can be a path to a ``folder`` or can be optionally of the structure ``fp:fn`` where ``fp``
-        is the path to the file and ``fn`` is the function name.
+      init_path (str): This can be a path to a `folder` or can be optionally of the structure `fp:fn` where `fp` is the path to the file and `fn` is the function name.
       project_name_or_id (str): The name or id of the LMAO project.
-      workspace_id (str, optional): If nor provided, defaults to global config
       trigger (bool, optional): Defaults to False. If True, will trigger the run after uploading.
       untracked (bool, optional): If True, then untracked files below 1MB will also be zipped and uploaded. Defaults to False.
       untracked_no_limit (bool, optional): If True, then all untracked files will also be zipped and uploaded. Defaults to False.
@@ -660,13 +560,14 @@ class LmaoCLI:
       resource_gpu_count (str, optional): Defaults to "0". Number of GPUs allocated to the experiment.
       resource_timeout (int, optional): Defaults to 120_000. The timeout between two consecutive runs, honoured but not guaranteed.
       resource_max_retries (int, optional): Defaults to 2. The maximum number of retries for a run.
-      **run_kwargs: These are the kwargs that will be passed to your Operator.
+      save_to_relic (bool, optional): Defaults to True. If True, will save the files to a relic when lmao.save_file() is called.
+      enable_system_monitoring (bool, optional): Defaults to False. If True, will enable system monitoring.
+      **run_kwargs: These are the kwargs that will be passed to your function.
     """
     # reconstruct the entire CLI command so we can show it in the UI
-    workspace_id = workspace_id or secret.get(ConfigString.workspace_id)
+    workspace_id = secret(AuthConfig.workspace_id)
     reconstructed_cli_comm = (
       f"nbx lmao upload '{init_path}' '{project_name_or_id}'"
-      f" --workspace_id '{workspace_id}'"
       f" --resource_cpu '{resource_cpu}'"
       f" --resource_memory '{resource_memory}'"
       f" --resource_disk_size '{resource_disk_size}'"
@@ -686,11 +587,15 @@ class LmaoCLI:
     if enable_system_monitoring:
       reconstructed_cli_comm += " --enable_system_monitoring"
 
+    if resource_max_retries < 1:
+      logger.error(f"max_retries must be >= 1. Got: {resource_max_retries}\n  Fix: set --max_retries=2")
+      raise ValueError()
+
     # uncommenting these since it can contain sensitive information, kept just in case for taking dictionaries as input
     # if relics_kwargs:
     #   rks = str(relics_kwargs).replace("'", "")
     #   reconstructed_cli_comm += f" --relics_kwargs '{rks}'"
-    
+
     for k, v in run_kwargs.items():
       if type(v) == bool:
         reconstructed_cli_comm += f" --{k}"
@@ -698,29 +603,23 @@ class LmaoCLI:
         reconstructed_cli_comm += f" --{k} '{v}'"
     logger.debug(f"command: {reconstructed_cli_comm}")
 
-    # clean user args
     resource_gpu_count = str(resource_gpu_count)
     if untracked_no_limit and not untracked:
       logger.debug("untracked_no_limit is True but untracked is False. Setting untracked to True")
       untracked = True
-    resource = Resource(
-      cpu = resource_cpu,
-      memory = resource_memory,
-      disk_size = resource_disk_size,
-      gpu = resource_gpu,
-      gpu_count = resource_gpu_count,
-      timeout = resource_timeout,
-      max_retries = resource_max_retries,
-    )
 
-    if resource.max_retries < 1:
-      logger.error(f"max_retries must be >= 1. Got: {resource.max_retries}\n  Fix: set --max_retries=2")
-      raise ValueError()
-
-    # first step is to get all the relevant information from the DB
-    workspace_id = workspace_id or secret.get(ConfigString.workspace_id)
+    # first step is to get all the relevant information from the DB and servers
+    workspace_id = workspace_id or secret(AuthConfig.workspace_id)
     project_name, project_id = get_project_name_id(project_name_or_id, workspace_id)
+    job_name = "nbxj_" + project_name[:15]
+    try:
+      job = Job(job_name = job_name)
+    except:
+      logger.warn(f"Job not found. Creating a new job: {job_name}")
+      job = Job.new(job_name = job_name, description = "automatically created by nbx LMAO for project: " + project_name)
+
     logger.info(f"Project: {project_name} ({project_id})")
+    lmao_stub = get_lmao_stub()
 
     # create a call init run and get the experiment metadata
     init_folder, _ = os.path.split(init_path)
@@ -729,28 +628,34 @@ class LmaoCLI:
       git_det = get_git_details(init_folder)
     else:
       git_det = {}
-    _metadata = {
-      "user_config": run_kwargs,
-      "git": git_det,
-      "resource": message_to_dict(resource),
-      "cli": reconstructed_cli_comm,
-      "lmao": {
-        "save_to_relic": save_to_relic,
-        "enable_system_monitoring": enable_system_monitoring,
-      }
-    }
 
-    job = Job("nbxj_" + project_name[:15], workspace_id = workspace_id)
-    lmao_stub = get_lmao_stub(secret.get("username"), workspace_id)
     run = lmao_stub.init_run(InitRunRequest(
+      workspace_id = workspace_id,
+      project_name = project_name,
+      project_id = project_id,
       agent_details = AgentDetails(
         type = AgentDetails.NBX.JOB,
         nbx_job_id = job.id, # run id will have to be updated from the job
       ),
-      created_at = SimplerTimes.get_now_i64(),
-      config = dumps(_metadata),
-      project_id = project_id,
+      config = ExperimentConfig(
+        run_kwargs = run_kwargs,
+        git = git_det,
+        resource = Resource(
+          cpu = str(resource_cpu),
+          memory = str(resource_memory),
+          disk_size = str(resource_disk_size),
+          gpu = str(resource_gpu),
+          gpu_count = str(resource_gpu_count),
+          timeout = int(resource_timeout),
+          max_retries = int(resource_max_retries),
+        ),
+        cli_comm = reconstructed_cli_comm,
+        save_to_relic = save_to_relic,
+        enable_system_monitoring = enable_system_monitoring,
+      ).to_json(),
     ))
+    if not project_id:
+      logger.info(f"Project {project_name} created with id: {run.project_id}")
     logger.info(f"Run ID: {run.experiment_id}")
 
     # connect to the relic
@@ -760,7 +665,7 @@ class LmaoCLI:
     if extra_keys:
       logger.error("Unknown arguments found:\n  * " + "\n  * ".join(extra_keys))
       raise RuntimeError("Unknown arguments found in the Relic")
-    relic = RelicsNBX(LMAO_RELIC_NAME, workspace_id = workspace_id, create = True, **relics_kwargs)
+    relic = Relics(LMAO_RELIC_NAME, workspace_id = workspace_id, create = True, **relics_kwargs)
 
     # create a git patch and upload it to relics
     if git_det:
@@ -775,11 +680,14 @@ class LmaoCLI:
         untracked_files = git_det["untracked_files"] # what to do with these?
         if untracked_files:
           # see if any file is larger than 10MB and if so, warn the user
+          warn_once = False
           for f in untracked_files:
             if os.path.getsize(f) > 1e7 and not untracked_no_limit:
               logger.warning(f"File: {f} is larger than 10MB and will not be available in sync")
-              logger.warning("  Fix: use git to track the file, avoid large files")
+              logger.warning("  Fix: use git to track small files, avoid large files")
+              logger.warning("  Fix: nbox.Relics can be used to store large files")
               logger.warning("  Fix: use --untracked-no-limit to upload all files")
+              warn_once = True
               continue
             zip_file.write(f, arcname = f)
       relic.put_to(_zf, f"{project_name}/{run.experiment_id}/git.zip")
@@ -787,50 +695,31 @@ class LmaoCLI:
       os.remove(patch_file)
 
     # tell the server that this run is being scheduled so atleast the information is visible on the dashboard
-    job: Job = Job.upload(init_path, id_or_name = "nbxj_" + project_name[:15], _ret = True) # keep only 20 chars
-    job.job_proto.resource.CopyFrom(resource)
-    job_proto: JobProto = nbox_grpc_stub.UpdateJob(
-      UpdateJobRequest(
-        job = job.job_proto,
-        update_mask = FieldMask(paths = ["resource"]),
-      )
+    upload_job_folder(
+      "job",
+      init_folder = init_path,
+      id = job.id,
+
+      # pass along the resource requirements
+      resource_cpu = resource_cpu,
+      resource_memory = resource_memory,
+      resource_disk_size = resource_disk_size,
+      resource_gpu = resource_gpu,
+      resource_gpu_count = resource_gpu_count,
+      resource_timeout = resource_timeout,
+      resource_max_retries = resource_max_retries,
     )
 
     if trigger:
-      logger.debug(f"Running job '{job.name}' ({job.id})")
-
-      # create the serialisable config
-      _lmaoConfig.clear()
-      _lmaoConfig.set(
-        workspace_id = workspace_id,
-        project_name = project_name,
-        project_id = project_id,
-        experiment_id = run.experiment_id,
-        metadata = _metadata,
-        save_to_relic = save_to_relic,
-        enable_system_monitoring = enable_system_monitoring,
-        args = (),
-        kwargs = run_kwargs,
-      )
-      
-      # from pprint import pprint
-      # pprint(_lmaoConfig.kv)
-      # exit()
-
-      # put the items in the relic and create a tag from it
-      fp = f"{project_name}/{run.experiment_id}"
-      relic.put_object(fp+"/init.pkl", _lmaoConfig.kv)
-      tag = f"{LMAO_JOB_TYPE_PREFIX}-{fp}"
-      logger.info(f"Run tag: {tag}")
+      fp = f"{run.project_id}/{run.experiment_id}"
+      tag = f"{LMAO_RM_PREFIX}{fp}"
+      logger.debug(f"Running job '{job.name}' ({job.id}) with tag: {tag}")
       job.trigger(tag)
 
-      # clear so the rest of the program doesn't get affected
-      _lmaoConfig.clear()
-
     # finally print the location of the run where the users can track this
-    logger.info(f"Run location: https://app.nimblebox.ai/workspace/{workspace_id}/monitoring/{project_id}/{run.experiment_id}")
+    logger.info(f"Run location: {secret(AuthConfig.url)}/workspace/{run.workspace_id}/monitoring/{run.project_id}/{run.experiment_id}")
 
 # do not change these it can become a huge pain later on
 LMAO_RELIC_NAME = "experiments"
-LMAO_JOB_TYPE_PREFIX = "NBXLmao"
-LMAO_ENV_VAR_PREFIX = "LMAO_"
+LMAO_RM_PREFIX = "NBXLmao-"
+LMAO_ENV_VAR_PREFIX = "NBX_LMAO_"
