@@ -1,3 +1,7 @@
+"""
+{% CallOut variant="info" label="This is currently a work in progress, please check back later for progress or reach out to NimbleBox support." /%}
+"""
+
 import time
 import psutil
 import threading
@@ -13,8 +17,8 @@ def get_metrics_dict():
   # total_memory = psutil.virtual_memory().total
   data = {
     "cpu_usage": sum(cpu_usage) / len(cpu_usage),
-    "memory_available (MiB)": psutil.virtual_memory().available // (1024 ** 2),
-    "memory_usage (MiB)": psutil.virtual_memory().used // (1024 ** 2),
+    "memory_available (MB)": psutil.virtual_memory().available // (1024 ** 2),
+    "memory_usage (MB)": psutil.virtual_memory().used // (1024 ** 2),
     "memory_percentage": psutil.virtual_memory().percent,
     "disk_utilisation": psutil.disk_usage('/').percent
   }
@@ -41,16 +45,16 @@ class SystemMetricsLogger:
         self._bar.wait()
         time.sleep(s)
     self.rl = threading.Thread(target=_rate_limiter, daemon=True)
-    self.rl.start()
-
-    # create a thread to create metrics
     self.metrics_logger = threading.Thread(target=self._create_metrics_dict, daemon=True)
-    self.metrics_logger.start()
 
     # So this function returns the latest cpu usage thast it has gathered from the previous call
     # and so we need to do an init fire to avoid getting 0.0
     psutil.cpu_percent(percpu=True)
 
+  def start(self):    
+    self.rl.start()
+    self.metrics_logger.start()
+
   def _create_metrics_dict(self):
     while True:
       if self.dk.completed:
@@ -73,12 +77,15 @@ class SystemMetricsLogger:
       self.dk.log(x, log_type=RunLog.LogType.SYSTEM)
     return False
 
-  def __del__(self):
+  def stop(self):
     self._bar.wait()
     self._bar.wait()
     self.rl.join()
     self.metrics_logger.join()
 
+  def __del__(self):
+    self.stop()
+
 
 if __name__ == "__main__":
   import json
