@@ -99,11 +99,14 @@ run in distributed fashion.
 - `@operator` decorator on your class and it will be run as a serving by default, you want that.
 
 ## Documentation
+
 """
 # Some parts of the code are based on the pytorch nn.Module class
 # pytorch license: https://github.com/pytorch/pytorch/blob/master/LICENSE
 # modifications: research@nimblebox.ai 2022
 
+from functools import partial
+import json
 import os
 import re
 import inspect
@@ -116,7 +119,7 @@ from concurrent.futures import ThreadPoolExecutor, as_completed
 from typing import Any, Dict, Iterable, List, Tuple, Union
 
 import nbox.utils as U
-from nbox.auth import secret, ConfigString
+from nbox.auth import secret, AuthConfig
 from nbox.utils import logger, SimplerTimes
 from nbox.nbxlib import operator_spec as ospec
 from nbox.nbxlib.tracer import Tracer
@@ -124,20 +127,21 @@ from nbox.version import __version__
 from nbox.sub_utils.latency import log_latency
 from nbox.framework.on_functions import get_nbx_flow
 from nbox.framework import AirflowMixin, PrefectMixin
-from nbox.hyperloop.job_pb2 import Job as JobProto, Resource
-from nbox.hyperloop.dag_pb2 import DAG, Flowchart, Node, RunStatus
-from nbox.network import _get_job_data
-from nbox.jobs import Schedule, Job, Serve
+from nbox.hyperloop.jobs.job_pb2 import Job as JobProto
+from nbox.hyperloop.common.common_pb2 import Resource
+from nbox.hyperloop.jobs.dag_pb2 import DAG, Flowchart, Node, RunStatus
+
+from nbox.jobs import Schedule, Job, Serve, _get_job_data
 from nbox.messages import write_binary_to_file
-from nbox.relics import RelicsNBX, RelicLocal
+from nbox.relics import Relics
 from nbox.init import nbox_ws_v1
 from nbox.subway import SpecSubway
 
+
 # alias
 DEFAULT_RESOURCE = ospec.DEFAULT_RESOURCE
 FN_IGNORE = ospec.FN_IGNORE
 
-
 class Operator():
   node = Node()
   source_edges: List[Node] = None
@@ -152,17 +156,13 @@ class Operator():
   # unless explicitly modified by `chief.machine.Machine`
   _current_resource = None
   _op_to_resource_map = {}
-  
-  # _all_keys = ['forward', 'node', 'source_edges', '_tracer', '_op_type', '_op_spec', '_current_resource', '_op_to_resource_map', '_operators']
 
   def __init__(self) -> None:
     """Create an operator, which abstracts your code into sharable, bulding blocks which
     can then deployed on either NBX-Jobs or NBX-Deploy.
     
-    Usage:
+    Examples:
 
-    .. code-block:: python
-      
       class MyOperator(Operator):
         def __init__(self, ...):
           ... # initialisation of job happens here
@@ -194,7 +194,7 @@ class Operator():
   def __remote_init__(self):
     """User can overwrite this function, this will be called only when running on remote.
     This helps in with things like creating the models can caching them in self, instead
-    of ``lru_cache`` in forward."""
+    of `lru_cache` in forward."""
     pass
 
   def remote_init(self):
@@ -257,31 +257,49 @@ class Operator():
   # 3. from a function, where we decorate an existing function and convert it to an operator
 
   @classmethod
-  def from_job(cls, job_id_or_name, workspace_id: str = ""):
-    """latch an existing job so that it can be called as an operator."""
+  def from_job(cls, job_name: str = "", job_id: str = "", workspace_id: str = ""):
+    """latch an existing job so that it can be called as an operator. This is designed to be used as
+    a part of "Compute Fabric" mode.
+
+    Args:
+      job_name (str, optional): Name of the job. Defaults to "".
+      job_id (str, optional): ID of the job. Defaults to "".
+    """
     # implement this when we have the client-server that allows us to get the metadata for the job
-    workspace_id = workspace_id or secret.get(ConfigString.workspace_id)
+    if (not job_name and not job_id) or (job_name and job_id):
+      raise ValueError("Either job_name or job_id must be specified")
+    workspace_id = workspace_id or secret(AuthConfig.workspace_id)
     if not workspace_id:
-      raise DeprecationWarning("Personal workspace does not support serving")
-    job_id, job_name = _get_job_data(job_id_or_name, workspace_id = workspace_id)
+      raise DeprecationWarning("workspace_id cannot be none. Reauth yourself with `nbx login`")
+    job_id, job_name = _get_job_data(
+      name = job_name,
+      id = job_id,
+      remove_archived = True, 
+      workspace_id = workspace_id
+    )
     if job_id is None:
       raise ValueError(f"No serving found with name {job_name}")
-    job = Job(job_id, workspace_id = workspace_id)
+    job = Job(job_id = job_id, workspace_id = workspace_id)
     logger.debug(f"Latching to job '{job_name}' ({job_id})")
-    
+
     def forward(*args, _wait: bool = True, **kwargs):
       """This is the forward method for a NBX-Job. All the parameters will be passed through Relics."""
-      logger.debug(f"Running job '{job_name}' ({job_id})")
-      relic = RelicsNBX("cache", workspace_id, create = True)
+      from nbox.lib.dist import RAW_DIST_RM_PREFIX, RAW_DIST_RELIC_NAME
+
+      run_unk = U.get_random_name(True).split("-")[0]
+      tag = f"{RAW_DIST_RM_PREFIX}{run_unk}"
+      logger.debug(f"Running job '{job_name}' ({job_id}): tag = {tag}")
 
       # determining the put location is very tricky because there is no way to create a sync between the
       # key put here and what the run will pull. This can lead to many weird race conditions. So for now
-      # I am going to rely on the fact that we cannot have two parallel active runs. Thus at any given
-      # moment there can be only one file at /{job_id}/args_kwargs.pkl
-      tag = U.get_random_name(True).split("-")[0]
-      relic.put_object(f"{job_id}/args_kwargs_{tag}", (args, kwargs))
+      # I am going to rely on the fact that we cannot have two parallel runs with same tag. Thus at any given
+      # moment there can be only one file at /{job_id}/args_kwargs_{tag}.pkl
+      relic = Relics(RAW_DIST_RELIC_NAME, create = True)
+      _pkl_id_in = run_unk + "_in"
+      _pkl_id_out = run_unk + "_out"
+      relic.put_object(_pkl_id_in, (args, kwargs))
 
-      # and then we will trigger the job and wait for the run to complete
+      # and then we will trigger the job with that specific tag and wait for the run to complete
       job.trigger(tag = tag)
       latest_run = job.last_n_runs(1)
 
@@ -289,11 +307,10 @@ class Operator():
         # even though we are returning the run_id, this cannot be trusted when making parallel call because of
         # the race condition on top run, which is the case when .map() method is called on the operator. 
         # return tag, latest_run["id"]
-        return tag, latest_run["id"]
+        return tag, latest_run["id"], _pkl_id_in, _pkl_id_out
 
       # if required wait else just return the 
       max_retries = latest_run['resource']['max_retries']
-
       max_polls = 600  # about 10 mins
       poll_count = 0
       while latest_run["retry_count"] <= max_retries:
@@ -352,7 +369,7 @@ class Operator():
       # this is deployment on a Build instance, there's a catch though without knowing the
       session.headers.update({
         "NBX-TOKEN": token,
-        "X-NBX-USERNAME": secret.get("username"),
+        "X-NBX-USERNAME": secret("username"),
       })
       serving_id = url
     else:
@@ -440,7 +457,7 @@ class Operator():
 
       # now we can call the function
       data = serving_stub.u(fn)(**_data)
-      
+
       # convert to usable values
       if not data["success"]:
         raise Exception(data["message"])
@@ -458,6 +475,7 @@ class Operator():
     # create the class and override some values to make more sense
     _op = cls()
     _op.__qualname__ = "serving_" + data["name"]
+
     _op.forward = forward
     _op._op_type = ospec.OperatorType.SERVING
     _op._op_spec = ospec._ServingSpec(
@@ -470,7 +488,7 @@ class Operator():
 
   @classmethod
   def fn(cls):
-    """Wraps the function as an Operator, so you can use all the same methods as Operator"""
+    """Wraps a function or class as an Operator, so you can use all the same methods as Operator"""
     def wrap(fn):
       if type(fn) == type(wrap): # lol the quick hack
         # this is a wrapped function to be run as a job
@@ -489,7 +507,7 @@ class Operator():
     return wrap
 
   def _cls(self, fn, *args, **kwargs):
-    """Do not use directly, use ``@operator`` decorator instead. Utility to wrap a class as an operator"""
+    """Do not use directly, use `@operator` decorator instead. Utility to wrap a class as an operator"""
     obj = fn(*args, **kwargs)
     # override the file this is necessary for getting the remote operations running
     self.__file__ = inspect.getfile(fn)
@@ -501,10 +519,26 @@ class Operator():
       wrap_obj = obj,
       init_ak = (args, kwargs)
     )
+
     return self
 
+  @classmethod
+  def from_class(cls, obj):
+    """Wraps an initialised class as an operator, so you can use all the same methods as Operator"""
+    op = cls()
+    op.__file__ = inspect.getfile(obj.__class__)
+    op.__doc__ = obj.__doc__
+    op.__qualname__ = "cls_" + obj.__class__.__qualname__
+    op._op_type = ospec.OperatorType.WRAP_CLS
+    op._op_spec = ospec._WrapClsSpec(
+      cls_name = obj.__class__.__qualname__,
+      wrap_obj = obj,
+      init_ak = ((), {})
+    )
+    return op
+
   def _fn(self, fn):
-    """Do not use directly, use ``@operator`` decorator instead. Utility to wrap a function as an operator"""
+    """Do not use directly, use `@operator` decorator instead. Utility to wrap a function as an operator"""
     self.forward = fn # override the forward function
     # override the file this is necessary for getting the remote operations running
     self.__file__ = inspect.getfile(fn)
@@ -512,8 +546,17 @@ class Operator():
     self.__qualname__ = "fn_" + fn.__name__
     self._op_type = ospec.OperatorType.WRAP_FN
     self._op_spec = ospec._WrapFnSpec(fn_name = fn.__name__, wrap_obj = fn)
+
     return self
 
+  @classmethod
+  def from_fn(cls, fn):
+    """Wraps a function as an operator, so you can use all the same methods as Operator"""
+    op = cls()
+    op = op._fn(fn)
+    return op
+    
+
   # /mixin
 
   # python state modification /
@@ -673,7 +716,6 @@ class Operator():
       self._tracer(self.node)
 
     # if user has enabled _tracking, then we will store the input, output values as well
-    
     return out
 
   def forward(self):
@@ -716,29 +758,53 @@ class Operator():
     dag.flowchart.CopyFrom(Flowchart(nodes = _nodes, edges = _edges))
     return dag
 
+  # def deploy_as_job(
+  #   self,
+  #   job_id: str = "",
+  #   **kwargs,
+  # ):
+  #   return self.deploy(
+  #     group_id = job_id,
+  #     deployment_type = ospec.OperatorType.JOB,
+  #     **kwargs,
+  #   )
+
+  # def deploy_as_serving(
+  #   self,
+  #   serving_id: str = "",
+  #   **kwargs,
+  # ):
+  #   return self.deploy(
+  #     group_id = serving_id,
+  #     deployment_type = ospec.OperatorType.SERVING,
+  #     **kwargs,
+  #   )
+
   def deploy(
     self,
-    workspace_id: str,
-    id_or_name:str = None,
+    group_id: str,
     deployment_type: str = None,
     resource: Resource = None,
     ignore_patterns: List[str] = [],
+    workspace_id: str = "",
     *,
     _unittest = False,
     _include_pattern = [],
   ) -> 'Operator':
     """Uploads relevant files to the cloud and deploys as a batch process or and API endpoint, returns the relevant
-    ``.from_job()`` or ``.from_serving`` Operator. This uploads the entire folder where the caller file is located.
-    In which case having a ``.nboxignore`` and ``requirements.txt`` will also be moved over.
+    `.from_job()` or `.from_serving` Operator. This uploads the entire folder where the caller file is located.
+    In which case having a `.nboxignore` and `requirements.txt` will be honoured.
 
     Args:
-      workspace_id (str): The workspace id to deploy to.
-      id_or_name (str, optional): The id or name of the deployment. if deployment_type is 'serving' this this must be provided.
-      deployment_type (str, optional): Defaults to 'serving' if WRAP_CLS else 'job'. The type of deployment to create.
-      resource (Resource, optional): The resource to deploy to, uses a reasonable default.
+      group_id (str): The Job/Deploy id to deploy to.
+      deployment_type (str, optional): The deployment type. Defaults to None.
+      resource (Resource, optional): The resource to use for deployment. Defaults to None.
+      ignore_patterns (List[str], optional): The patterns to ignore. Defaults to [].
     """
+
     if not workspace_id:
-      raise ValueError("Must provide a workspace_id")
+      workspace_id = secret(AuthConfig.workspace_id)
+      logger.info(f"Using workspace_id: {workspace_id}")
 
     # go over reasonable checks for deployment
     if deployment_type == None:
@@ -750,20 +816,17 @@ class Operator():
       raise ValueError(f"Cannot deploy a class as a job, only as a serving")
     if deployment_type not in ospec.OperatorType._valid_deployment_types():
       raise ValueError(f"Invalid deployment type: {deployment_type}. Must be one of {ospec.OperatorType._valid_deployment_types()}")
-    if deployment_type == ospec.OperatorType.SERVING.value:
-      if id_or_name is None:
-        raise ValueError("id_or_name must be provided for serving deployment")
+    if deployment_type == ospec.OperatorType.SERVING:
+      if group_id is None:
+        raise ValueError("pass deployment_id as group_id for serving deployment")
 
     # get the filepath and name to import for convience
     fp, folder, file, name = ospec.get_operator_location(self)
     logger.info(f"Deployment Type: {deployment_type}")
-    logger.info(f"Deploying '{name}' from '{fp}'")
-    logger.info(f"Will upload folder: {folder}")
 
     # create a temporary directory to store all the files
     # copy over all the files and wait for it, else the changes below won't be reflected
     # dude damn, how did I not know this: https://dev.to/ackshaey/macos-vs-linux-the-cp-command-will-trip-you-up-2p00
-
     # so basically till now all we know is that `from fp import name` and `name()`. This process can now
     # be used to create a custom `nbx_user.py` file from which `exe.py` can import the three functions
     # and use them the conventional way. the three functions could simply be a router and the real operator
@@ -775,7 +838,7 @@ class Operator():
     with open(U.join(folder, "nbx_user.py"), "w") as f:
       f.write(f'''# Autogenerated for .deploy() call
 from nbox.messages import read_file_to_binary
-from nbox.hyperloop.job_pb2 import Resource
+from nbox.hyperloop.jobs.job_pb2 import Resource
 
 from {file.split('.')[0]} import {name}
 
@@ -813,19 +876,25 @@ get_schedule = lambda: None
     if _unittest:
       return
 
-    if deployment_type == ospec.OperatorType.JOB.value:
+    init_folder = f"{fp.replace('.py', '')}:{name}"
+    logger.debug(f"Upload init_folder: {init_folder}")
+
+    if deployment_type == ospec.OperatorType.JOB:
       out = Job.upload(
-        init_folder = folder,
-        id_or_name = id_or_name or "dot_deploy_runs", # all the .deploy() methods will be called dot_deploy_runs
-        workspace_id = workspace_id,
+        init_folder = init_folder,
+        id = group_id,
+        _ret = True
+      )
+      return self.from_job(job_id = out.id)
+    elif deployment_type == ospec.OperatorType.SERVING:
+      out = Serve.upload(
+        init_folder = init_folder,
+        id = group_id,
         _ret = True
       )
-      return self.from_job(job_id_or_name = out.id, workspace_id = workspace_id)
-    elif deployment_type == ospec.OperatorType.SERVING.value:
-      out = Serve.upload(init_folder = folder, id_or_name = id_or_name, workspace_id = workspace_id, _ret = True)
 
       # get the serving object
-      stub = nbox_ws_v1.workspace.u(workspace_id).deployments.u(out.id)
+      stub = nbox_ws_v1.deployments.u(out.id)
 
       # we will poll here till the model is not ready and return the latched operator
       done = False
@@ -946,10 +1015,13 @@ if __name__ == "__main__":
       return results
 
     elif self._op_type == ospec.OperatorType.JOB:
+      from nbox.lib.dist import RAW_DIST_RELIC_NAME
+
       # now this is simple enough, we need to implement a waiting queue that's it
       _start_time = monotonic()
       run_tags = []
       run_tag_to_input_idx = {}
+      run_tag_to_out_files = {}
 
       # though the threadpool is a faster method, cannot trust the run_id because of the race condition on top run
       with ThreadPoolExecutor(min(len(inputs), 10)) as exe:
@@ -957,9 +1029,10 @@ if __name__ == "__main__":
         results = {exe.submit(trigger, i, x) for i,x in enumerate(inputs)}
         for future in as_completed(results):
           try:
-            i, (tag, run_id) = future.result()
+            (i, tag, _pkl_id_in, _pkl_id_out) = future.result()
             run_tags.append(tag)
             run_tag_to_input_idx[tag] = i
+            run_tag_to_out_files[tag] = _pkl_id_out
           except Exception as e:
             raise Exception(e)
 
@@ -988,15 +1061,14 @@ if __name__ == "__main__":
       run_status = {rid: None for rid in run_ids}
       active_runs = {rid for rid, done in run_status.items() if not done}
       pbar = tqdm(total = len(inputs), desc = f"Waiting for runs ({html_path})")
-      relic = RelicsNBX("cache", workspace_id = self._op_spec.workspace_id)
+      relic = Relics(RAW_DIST_RELIC_NAME, workspace_id = self._op_spec.workspace_id)
 
       def _update_runs():
         # get all runs, filter those in this invocation, update the status
         files_ = {}
-        for t in run_tags:
-          files_[t] = (
-            relic.has(f"{self._op_spec.job_id}/return_{tag}"),
-          )
+        for t, _pkl_id_out in run_tag_to_out_files.items():
+          files_[t] = relic.has(_pkl_id_out)
+
         out = self._op_spec.job.last_n_runs(len(inputs))
         # print(out)
         if not isinstance(out, (list, tuple)):
